"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const os_1 = __importDefault(require("os"));
const path_1 = __importDefault(require("path"));
const process_1 = __importDefault(require("process"));
const tmp_1 = __importDefault(require("tmp"));
const chalk_1 = __importDefault(require("chalk"));
const cross_spawn_1 = __importDefault(require("cross-spawn"));
const express_1 = __importDefault(require("express"));
const serve_index_1 = __importDefault(require("serve-index"));
const express_ws_1 = __importDefault(require("express-ws"));
const ws_1 = __importDefault(require("ws"));
const glob_1 = __importDefault(require("glob"));
const chokidar_1 = __importDefault(require("chokidar"));
const open_1 = __importDefault(require("open"));
const util_1 = __importDefault(require("util"));
const readFileAsync = util_1.default.promisify(fs_1.default.readFile);
const globAsync = util_1.default.promisify(glob_1.default);
const statAsync = util_1.default.promisify(fs_1.default.stat);
tmp_1.default.setGracefulCleanup();
express_1.default.static.mime.define({ "text/plain; charset=UTF-8": ["elm"] });
express_1.default.static.mime.define({ "text/plain; charset=UTF-8": ["md"] });
const npmPackage = require(path_1.default.join(__dirname, "../package.json"));
function info(...args) {
    console.log(...args);
}
function warning(...args) {
    console.log(chalk_1.default.yellow(...args));
}
function error(...args) {
    console.log(chalk_1.default.red(...args));
}
function fatal(...args) {
    error(...args);
    chalk_1.default.red("Exiting...");
    process_1.default.exit(1);
}
function elmErrors(error) {
    console.log(elmErrorWithColor(error.errors));
}
const elmErrorWithColor = (errors) => {
    const repeat = (str, num, min = 3) => [...Array(num < 0 ? min : num)].map(_ => str).join('');
    const errorToString = (error) => {
        const problemToString = (problem) => {
            // Removing the elm-stuff/generatedFolderName from the beginning of the filepath
            let errorPath = error.path.substr(process_1.default.cwd().length + 1).split(path_1.default.sep);
            errorPath.shift();
            errorPath.shift();
            const errorFilePath = errorPath.join(path_1.default.sep);
            return [
                chalk_1.default.cyan(`-- ${problem.title} ${repeat('-', 63 - problem.title.length - errorFilePath.length)} ${errorFilePath}`),
                problem.message.map(messageToString).join('')
            ].join('\n\n');
        };
        const messageToString = (line) => {
            if (typeof line === 'string') {
                return line;
            }
            else {
                let message = line.string;
                if (line.bold) {
                    message = chalk_1.default.bold(message);
                }
                if (line.underline) {
                    message = chalk_1.default.underline(message);
                }
                switch (line.color) {
                    case 'green':
                        message = chalk_1.default.green(message);
                        break;
                    case 'yellow':
                        message = chalk_1.default.yellow(message);
                        break;
                    case 'cyan':
                        message = chalk_1.default.cyan(message);
                        break;
                    case 'RED':
                        message = chalk_1.default.red(message);
                        break;
                    default:
                        break;
                }
                return message;
            }
        };
        return error.problems.map(problemToString).join('\n\n');
    };
    return errors.map(errorToString).join('\n\n\n');
};
/*
 * Find and check Elm executable
 */
function getElm() {
    let elm = (args, cwd = ".") => cross_spawn_1.default.sync("npx", ["--no-install", "elm"].concat(args), { cwd });
    let exec = elm(["--version"]);
    if (exec.error || exec.status !== 0 || exec.stderr.toString().length > 0) {
        elm = (args, cwd = ".") => cross_spawn_1.default.sync("elm", args, { cwd });
        exec = elm(["--version"]);
    }
    if (exec.error) {
        fatal(`cannot run 'elm --version' (${exec.error})`);
    }
    else if (exec.status !== 0) {
        error(`cannot run 'elm --version':`);
        process_1.default.stderr.write(exec.stderr);
        process_1.default.exit(exec.status ? exec.status : 1);
    }
    const version = exec.stdout.toString().trim();
    if (!version.startsWith("0.19")) {
        fatal(`unsupported Elm version ${version}`);
    }
    return [elm, version];
}
function getElmCache(elmVersion) {
    const dir = os_1.default.platform() === "win32" ? "AppData/Roaming/elm" : ".elm";
    const home = process_1.default.env.ELM_HOME || path_1.default.join(os_1.default.homedir(), dir);
    const packages = elmVersion === "0.19.0" ? "package" : "packages";
    const cache = path_1.default.join(home, elmVersion, packages);
    return cache;
}
async function getManifest(manifestPath) {
    return readFileAsync(manifestPath, "utf8")
        .then(async (json) => {
        let manifest = JSON.parse(json);
        let stat = await statAsync(manifestPath);
        manifest["timestamp"] = Math.round(stat.mtime.getTime() / 1000);
        manifest = completeApplication(manifestPath, manifest);
        return manifest;
    })
        .catch(err => error(err));
}
function getManifestSync(manifestPath) {
    try {
        const json = fs_1.default.readFileSync(manifestPath, "utf8");
        let manifest = JSON.parse(json);
        let stat = fs_1.default.statSync(manifestPath);
        manifest["timestamp"] = Math.round(stat.mtime.getTime() / 1000);
        return completeApplication(manifestPath, manifest);
    }
    catch (err) {
        return null;
    }
}
function completeApplication(manifestPath, manifest) {
    if (manifest.type !== "application") {
        return manifest;
    }
    try {
        const elmAppPath = path_1.default.resolve(path_1.default.dirname(manifestPath), "elm-application.json");
        if (fs_1.default.existsSync(elmAppPath)) {
            const elmApp = JSON.parse(fs_1.default.readFileSync(elmAppPath).toString());
            Object.assign(manifest, elmApp);
        }
    }
    catch (err) {
        error(err);
    }
    if (!("name" in manifest)) {
        manifest.name = "my/application";
    }
    if (!("version" in manifest)) {
        manifest.version = "1.0.0";
    }
    if (!("summary" in manifest)) {
        manifest.summary = "Elm application";
    }
    if (!("license" in manifest)) {
        manifest.license = "Fair";
    }
    return manifest;
}
function fullname(manifest) {
    return `${manifest.name}/${manifest.version}`;
}
async function searchPackages(pattern) {
    let paths = await globAsync(pattern + "/elm.json", { realpath: true });
    const manifests = await Promise.all(paths.map(path => getManifest(path)));
    let packages = manifests.reduce((acc, pkg) => {
        if (pkg.name && pkg.name in acc && pkg.version) {
            acc[pkg.name].versions.push(pkg.version);
            return acc;
        }
        else if (pkg.name && pkg.version) {
            acc[pkg.name] = {
                name: pkg.name,
                summary: pkg.summary || "",
                license: pkg.license || "Fair",
                versions: [pkg.version]
            };
            return acc;
        }
        else {
            warning("invalid elm.json", pkg);
            return acc;
        }
    }, {});
    return packages;
}
async function packageReleases(pattern) {
    const paths = await globAsync(pattern + "/elm.json", {
        realpath: true
    });
    const manifests = await Promise.all(paths.map(path => getManifest(path)));
    let releases = manifests.reduce((releases, pkg) => {
        if (pkg.version && pkg.timestamp) {
            releases[pkg.version] = pkg.timestamp;
            return releases;
        }
        else {
            return releases;
        }
    }, {});
    return releases;
}
function merge(objects) {
    return objects.reduce((acc, obj) => Object.assign(acc, obj));
}
function buildDocs(manifest, dir, elm, clean = true, verbose = false) {
    info(`  |> building ${path_1.default.resolve(dir)} documentation`);
    try {
        if (manifest.type == "package") {
            return buildPackageDocs(dir, elm, clean, verbose);
        }
        else if (manifest.type == "application") {
            return buildApplicationDocs(manifest, dir, elm, clean, verbose);
        }
    }
    catch (err) {
        error(err);
    }
    return {};
}
// Return a docs.json or a json error report
function buildPackageDocs(dir, elm, clean, verbose) {
    const tmpFile = tmp_1.default.fileSync({ prefix: "elm-docs-", postfix: ".json" });
    const buildDir = path_1.default.resolve(dir);
    if (!clean) {
        info(`  |> generating ${tmpFile.name} documentation`);
    }
    const build = elm(["make", `--docs=${tmpFile.name}`, "--report=json"], buildDir);
    if (build.error) {
        error(`cannot build documentation (${build.error})`);
    }
    else if (build.stderr.toString().length > 0 && verbose) {
        elmErrors(JSON.parse(build.stderr.toString()));
    }
    let docs;
    try {
        docs = JSON.parse(fs_1.default.readFileSync(tmpFile.name).toString());
    }
    catch (err) {
        try {
            // Return Errors JSON report
            docs = JSON.parse(build.stderr.toString());
        }
        catch (err) {
            docs = {};
        }
    }
    if (clean) {
        tmpFile.removeCallback();
    }
    return docs;
}
function buildApplicationDocs(manifest, dir, elm, clean, verbose) {
    // Build package from application manifest
    const elmStuff = path_1.default.resolve(dir, "elm-stuff");
    if (!fs_1.default.existsSync(elmStuff)) {
        fs_1.default.mkdirSync(elmStuff);
    }
    const tmpDir = tmp_1.default.dirSync({
        dir: elmStuff,
        prefix: "elm-application-",
        unsafeCleanup: true
    });
    const tmpDirSrc = path_1.default.resolve(tmpDir.name, "src");
    fs_1.default.mkdirSync(tmpDirSrc);
    if (!clean) {
        info(`  |> generating ${tmpDir.name} package`);
    }
    // Build package manifest
    let pkg = {
        type: "package",
        name: manifest.name,
        summary: manifest.summary,
        license: manifest.license,
        version: manifest.version,
        "exposed-modules": manifest["exposed-modules"],
        "elm-version": versionToConstraint(manifest["elm-version"]),
        dependencies: {},
        "test-dependencies": {},
        timestamp: manifest.timestamp
    };
    // Add dependencies constraints
    if (manifest.dependencies.direct) {
        for (const [name, version] of Object.entries(manifest.dependencies.direct)) {
            pkg.dependencies[name] = versionToConstraint(version);
        }
    }
    // Add source directories exposed-modules
    let exposedModules = getExposedModules(pkg["exposed-modules"]);
    if (manifest["source-directories"]) {
        manifest["source-directories"].forEach(src => {
            const srcDir = path_1.default.resolve(src);
            importModules(srcDir, tmpDirSrc);
            const elmJsonPath = path_1.default.resolve(src, "../elm.json");
            if (fs_1.default.existsSync(elmJsonPath)) {
                try {
                    const srcManifest = getManifestSync(elmJsonPath);
                    if (srcManifest && srcManifest.type === "package") {
                        const srcModules = getExposedModules(srcManifest["exposed-modules"]);
                        exposedModules = exposedModules.concat(srcModules);
                    }
                }
                catch (err) {
                    error(err);
                }
            }
        });
    }
    pkg["exposed-modules"] = exposedModules;
    // Write elm.json and generate package documentation
    const elmJson = JSON.stringify(pkg);
    fs_1.default.writeFileSync(tmpDir.name + "/elm.json", elmJson, "utf8");
    const docs = buildPackageDocs(tmpDir.name, elm, clean, verbose);
    // remove temporary directory
    if (clean) {
        tmpDir.removeCallback();
    }
    return docs;
}
function getExposedModules(manifestExposedModules) {
    let exposedModules = [];
    if (manifestExposedModules) {
        if (Array.isArray(manifestExposedModules)) {
            exposedModules = manifestExposedModules;
        }
        else if (typeof manifestExposedModules === "object") {
            Object.values(manifestExposedModules).forEach(modules => {
                exposedModules = exposedModules.concat(modules);
            });
        }
    }
    return exposedModules;
}
function importModules(srcDir, dstDir) {
    glob_1.default.sync("**/*.elm", { cwd: srcDir }).forEach(elm => {
        try {
            const dir = path_1.default.resolve(dstDir, path_1.default.dirname(elm));
            mkdirSyncRecursive(path_1.default.resolve(dstDir, dir));
            const srcModulePath = path_1.default.resolve(srcDir, elm);
            const dstModulePath = path_1.default.resolve(dstDir, elm);
            let module = fs_1.default.readFileSync(srcModulePath).toString();
            if (module.match(/^port +module /) !== null) {
                // Stub ports by subscriptions and commands that do nothing
                info(`  |> stubbing ${elm} ports`);
                module = module.replace(/^port +([^ :]+)([^\n]+)$/mg, (match, name, decl, _off, _str) => {
                    if (name === "module") {
                        return ["module", decl].join(" ");
                    }
                    else if (decl.includes("Sub")) {
                        info("  |> stubbing incoming port", name);
                        return name + " " + decl + "\n" + name + " = always Sub.none\n";
                    }
                    else if (decl.includes("Cmd")) {
                        info("  |> stubbing outgoing port", name);
                        return name + " " + decl + "\n" + name + " = always Cmd.none\n";
                    }
                    else {
                        warning("unmatched", match);
                    }
                    return match;
                });
                fs_1.default.writeFileSync(dstModulePath, module);
            }
            else {
                linkModule(srcModulePath, dstModulePath);
            }
        }
        catch (err) {
            error(err);
        }
    });
}
function mkdirSyncRecursive(dir) {
    const absoluteDir = path_1.default.resolve(dir);
    const sep = path_1.default.sep;
    absoluteDir.split(sep).reduce((parent, child) => {
        const d = path_1.default.resolve(parent, child);
        try {
            if (!fs_1.default.existsSync(d)) {
                fs_1.default.mkdirSync(d);
            }
        }
        catch (err) {
            error(err);
        }
        return d;
    }, "/");
}
function linkModule(linked, link) {
    try {
        if (!fs_1.default.existsSync(link)) {
            if (os_1.default.platform() === "win32") {
                // Windows requires to be admin to create symlinks
                fs_1.default.copyFileSync(linked, link);
            }
            else {
                fs_1.default.symlinkSync(linked, link);
            }
        }
    }
    catch (err) {
        error(err);
    }
}
function versionToConstraint(version) {
    const [major, minor, patch] = version.split(".", 3);
    const nextPatch = parseInt(patch) + 1;
    return `${major}.${minor}.${patch} <= v < ${major}.${minor}.${nextPatch}`;
}
class DocServer {
    constructor(options) {
        const { dir = ".", port = 8000, browser = true, reload = true, debug = false } = options || {};
        this.options = {
            browser,
            debug,
            dir: fs_1.default.lstatSync(dir).isFile() ? path_1.default.dirname(dir) : path_1.default.resolve(dir),
            port,
            reload
        };
        try {
            process_1.default.chdir(this.options.dir);
        }
        catch (err) {
            error(err);
        }
        [this.elm, this.elmVersion] = getElm();
        this.elmCache = getElmCache(this.elmVersion);
        let app = express_1.default();
        this.ws = express_ws_1.default(app);
        this.app = this.ws.app;
        this.wss = this.ws.getWss();
        this.manifest = getManifestSync("elm.json");
        info(chalk_1.default `{bold elm-doc-preview ${npmPackage.version}} using elm ${this.elmVersion}`);
        if (this.manifest && this.manifest.name && this.manifest.version) {
            info(chalk_1.default `Previewing {magenta ${this.manifest.name} ${this.manifest.version}}`, `from ${this.options.dir}`);
        }
        else {
            info(`No package or application found in ${this.options.dir},`, "running documentation server only");
        }
        this.setupWebServer();
        if (this.manifest && this.options.reload) {
            this.setupFilesWatcher();
        }
    }
    setupWebServer() {
        this.app.use("/", express_1.default.static(path_1.default.join(__dirname, "../static"), {
            index: "../static/index.html"
        }));
        // websockets
        this.app.ws("/", (socket, req) => {
            info(`  |> ${req.connection.remoteAddress} connected`);
            socket.on("close", () => {
                info("  |> client disconnected");
            });
        });
        // preview
        this.app.get("/preview", (_req, res) => {
            if (this.manifest) {
                res.json(this.manifest);
            }
        });
        // search.json
        this.app.get("/search.json", (_req, res) => {
            Promise.all([`${this.elmCache}/*/*/*`, "."]
                .map(pattern => searchPackages(pattern)))
                .then(packagesArray => {
                // add/overwrite cache with project
                res.json(Object.values(merge(packagesArray)));
            })
                .catch(err => error(err));
        });
        // releases.json
        this.app.get("/packages/:author/:project/releases.json", (req, res) => {
            const p = req.params;
            const name = `${p.author}/${p.project}`;
            let dirs = [`${this.elmCache}/${name}/*`];
            if (this.manifest && this.manifest.name === name) {
                dirs.push(".");
            }
            Promise.all(dirs.map(dir => packageReleases(dir)))
                .then(releasesArray => {
                // add/overwrite cache with project
                res.json(merge(releasesArray));
            })
                .catch(err => error(err));
        });
        // docs.json
        this.app.get("/packages/:author/:project/:version/docs.json", (req, res) => {
            const p = req.params;
            const name = `${p.author}/${p.project}/${p.version}`;
            if (this.manifest && fullname(this.manifest) === name) {
                res.json(buildDocs(this.manifest, ".", this.elm, !this.options.debug));
            }
            else {
                res.sendFile(path_1.default.resolve(this.elmCache, name, "docs.json"));
            }
        });
        // Serve README.md files
        this.app.get("/packages/:author/:project/:version/README.md", (req, res) => {
            const p = req.params;
            const name = [p.author, p.project, p.version].join("/");
            if (this.manifest && fullname(this.manifest) === name) {
                res.sendFile(path_1.default.resolve(".", "README.md"));
            }
            else {
                res.sendFile(path_1.default.resolve(this.elmCache, name, "README.md"));
            }
        });
        // Serve elm.json files
        this.app.get("/packages/:author/:project/:version/elm.json", (req, res) => {
            const p = req.params;
            const name = `${p.author}/${p.project}/${p.version}`;
            if (this.manifest && fullname(this.manifest) === name) {
                const manifest = getManifestSync("elm.json");
                res.json(manifest);
            }
            else {
                res.sendFile(path_1.default.resolve(this.elmCache, name, "elm.json"));
            }
        });
        let setHeaders = (res, path, _stat) => {
            if (path.endsWith("/LICENSE")) {
                res.setHeader("Content-Type", "text/plain; charset=UTF-8");
            }
        };
        // Serve project source
        if (this.manifest) {
            this.app.use(`/source/${fullname(this.manifest)}`, express_1.default.static(".", { setHeaders: setHeaders }), serve_index_1.default(".", { icons: true }));
        }
        // Serve cached packages source
        this.app.use("/source", express_1.default.static(this.elmCache, { setHeaders: setHeaders }), serve_index_1.default(this.elmCache, { icons: true }));
        // default route
        this.app.get("*", (_req, res) => {
            res.sendFile(path_1.default.join(__dirname, "../static/index.html"));
        });
    }
    setupFilesWatcher() {
        // We use glob patterns to avoid https://github.com/paulmillr/chokidar/issues/237.
        // We want to watch ["elm.json", "elm-application.json", "README.md"].
        const glob = ["elm*.json", "README*.md"];
        if (this.manifest && this.manifest["source-directories"]) {
            this.manifest["source-directories"].forEach(src => {
                glob.push(src + "/**/*.elm");
                glob.push(path_1.default.normalize(src + "/../elm.json"));
            });
        }
        else if (this.manifest) {
            glob.push("src/**/*.elm");
        }
        const watcher = chokidar_1.default.watch(glob, {
            ignored: ["**/node_modules", "**/elm-stuff", "**/.git"],
            ignoreInitial: true,
            atomic: true
        });
        watcher
            .on("all", (_event, filepath) => this.onChange(filepath))
            .on("error", err => error(err))
            .on("ready", () => {
            if (this.manifest && this.manifest.type === "package") {
                info(`  |> watching package`);
            }
            else if (this.manifest && this.manifest.type === "application") {
                info(`  |> watching application`);
            }
            if (this.options.debug) {
                info(watcher.getWatched());
            }
        });
    }
    onChange(filepath) {
        info("  |>", "detected", filepath, "modification");
        if (filepath == "README.md") {
            this.sendReadme();
        }
        else if (filepath.endsWith(".json")) {
            this.manifest = getManifestSync("elm.json");
            this.sendManifest();
            this.sendDocs();
        }
        else {
            this.sendDocs();
        }
    }
    sendReadme() {
        const readme = path_1.default.join(this.options.dir, "README.md");
        if (this.manifest && this.manifest.name && this.manifest.name.includes("/")) {
            const [author, project] = this.manifest.name.split("/", 2);
            try {
                info("  |>", "sending README");
                this.broadcast({
                    type: "readme",
                    data: {
                        author: author,
                        project: project,
                        version: this.manifest.version,
                        readme: fs_1.default.readFileSync(readme).toString()
                    }
                });
            }
            catch (err) {
                error(err);
            }
        }
    }
    sendManifest() {
        if (this.manifest && this.manifest.name && this.manifest.name.includes("/")) {
            const [author, project] = this.manifest.name.split("/", 2);
            info("  |>", "sending Manifest");
            this.broadcast({
                type: "manifest",
                data: {
                    author: author,
                    project: project,
                    version: this.manifest.version,
                    manifest: this.manifest
                }
            });
        }
    }
    sendDocs() {
        if (this.manifest && this.manifest.name && this.manifest.name.includes("/")) {
            const docs = buildDocs(this.manifest, this.options.dir, this.elm, !this.options.debug);
            const [author, project] = this.manifest.name.split("/", 2);
            info("  |>", "sending Docs");
            this.broadcast({
                type: "docs",
                data: {
                    author: author,
                    project: project,
                    version: this.manifest.version,
                    time: this.manifest.timestamp,
                    docs: docs
                }
            });
        }
    }
    listen() {
        return this.app.listen(this.options.port, () => {
            if (this.options.browser && this.manifest && this.manifest.name && this.manifest.version) {
                open_1.default(`http://localhost:${this.options.port}/packages/${this.manifest.name}/${this.manifest.version}/`);
            }
            else if (this.options.browser) {
                open_1.default(`http://localhost:${this.options.port}`);
            }
            info(chalk_1.default `{blue Browse} {bold {green <http://localhost:${this.options.port.toString()}>}}`, chalk_1.default `{blue to see your documentation}`);
        });
    }
    make(filename) {
        if (this.manifest) {
            const docs = buildDocs(this.manifest, this.options.dir, this.elm, !this.options.debug, true);
            info(`  |> writing documentation into ${filename}`);
            if (Array.isArray(docs) && docs.length > 0) {
                if (filename !== "/dev/null") {
                    try {
                        fs_1.default.writeFileSync(filename, JSON.stringify(docs), "utf8");
                    }
                    catch (err) {
                        fatal(err);
                    }
                }
                process_1.default.exit(0);
            }
            else {
                fatal("failed to build project documentation");
            }
        }
    }
    broadcast(obj) {
        this.wss.clients.forEach(client => {
            if (client.readyState === ws_1.default.OPEN) {
                client.send(JSON.stringify(obj));
            }
        });
    }
}
module.exports = DocServer;
//# sourceMappingURL=elm-doc-server.js.map